/**
 * @file executor.c
 * @author Florian Schmitt <florian.schmitt@qikgrp.com>
 * ----------------------------------------------------------------------------
 * (c) Florian Schmitt / 2012-2013
 * License GPLv3+: GNU GPL Version 3 
 * or higher <http://gnu.org/licenses/gpl.html>
 * This code comes with ABSOLUTELY NO WARRANTY. This is free software, and 
 * you are welcome to redistribute it under certain conditions.
 * ----------------------------------------------------------------------------
 * Executes string commands
**/
#include "executor.h"
#include "dmxplayer.h"
#include "dmxprogram.h"
#include "dmxstr.h"
#include "cmdbuf.h"
#include "gui/listing.h"
#include "midi/midirecv.h"
#include <sys/stat.h>



/**
 * executor()
 * Executes a command
 * @param cmd      Command to execute
 * @param verbose  if zero no output is thrown (MUST BE ZERO FOR THREADS!)
**/
int executor(const char *cmd, int verbose)
{
	const char *str = 0;
	int d, e;

	/* Comment or empty line */
	if(dmxstr_is_comment(cmd))
	{
		return EXECUTOR_SUCCESS;
	}

	/* tap */
	if( (str = dmxstr_chkcmd(cmd, "tap")) != 0 )
	{
		d = dmxplayer_timetap();
		if(verbose) cmdbuf_printf("tap step: %d", (d+1));
		return EXECUTOR_SUCCESS;
	}

	/* sync */
	else if( (str = dmxstr_chkcmd(cmd, "sync")) != 0 )
	{
		dmxplayer_timesync();
		return EXECUTOR_SUCCESS;
	}

	/* clear (programmer) */
	else if( (str = dmxstr_chkcmd(cmd, "clear")) != 0 )
	{
		dmxplayer_clear_prog();
		return EXECUTOR_SUCCESS;
	}

	/* store */
	else if( (str = dmxstr_chkcmd(cmd, "store")) != 0 )
	{
		d = listing_getSelectedProgram();
		dmxplayer_store_program(d);
		return EXECUTOR_SUCCESS;
	}

	/* clear (screen) */
	else if( (str = dmxstr_chkcmd(cmd, "cls")) != 0 )
	{
		if(verbose) cmdbuf_clear();
		return EXECUTOR_SUCCESS;
	}

	/* setbpm */
	else if((str = dmxstr_chkcmd(cmd, "setbpm")) != 0 )
	{
		if( (str = dmxstr_get_next_param_int(str, &d)) != 0)
		{
			dmxplayer_setbpm(d);
			return EXECUTOR_SUCCESS;
		}
		if(verbose) cmdbuf_printf("usage: setbpm <number>");
		return EXECUTOR_FAILURE;
	}

	/* echo */
	else if((str = dmxstr_chkcmd(cmd, "echo")) != 0 )
	{
		str = dmxstr_remove_whitespace(str);
		if(verbose) cmdbuf_printf(str);
		return EXECUTOR_SUCCESS;
	}

	/* chaser */
	else if( ((str = dmxstr_chkcmd(cmd, "chaser")) != 0 ) ||
	         ((str = dmxstr_chkcmd(cmd, "chs"))    != 0 )    )
	{
		if( (str = dmxstr_get_next_param_int(str, &d)) != 0)
		{
			return executor_chaser(d, str, 0, verbose);
		}
		if(verbose) cmdbuf_printf("usage: chaser <number> <command>");
		return EXECUTOR_FAILURE;
	}

	/* channel */
	else if( ( (str = dmxstr_chkcmd(cmd, "channel")) != 0 ) ||
		      ( (str = dmxstr_chkcmd(cmd, "c"))       != 0 )    )
	{
		if( (str = dmxstr_get_next_param_range(str, &d, &e)) != 0)
		{
			return executor_channel(d, e, str, verbose);
		}
		if(verbose) cmdbuf_printf("usage: channel <number> <command>");
		return EXECUTOR_FAILURE;
	}

	/* exec - execute script files */
	else if( (str = dmxstr_chkcmd(cmd, "exec")) != 0 )
	{
		str = dmxstr_remove_whitespace(str);
		return executor_exec(str, verbose);
	}

	/* midibind - binds midi to a command */
	else if( ((str = dmxstr_chkcmd(cmd, "midibind")) != 0 ) ||
	         ((str = dmxstr_chkcmd(cmd, "mbind"))    != 0 )    )
	{
		str = dmxstr_remove_whitespace(str);
		return executor_midibind(str, verbose);
	}

	/* midiconnect - begins to listen on an interface*/
	else if( ((str = dmxstr_chkcmd(cmd, "midiconnect")) != 0 ) ||
	         ((str = dmxstr_chkcmd(cmd, "mcon"))    != 0 )    )
	{
		str = dmxstr_remove_whitespace(str);
		if( midireceiver_add_midi_device(str) )
			return EXECUTOR_SUCCESS;
		if(verbose) cmdbuf_printf("can not start device thread!");
		return EXECUTOR_FAILURE;
	}

	/* try external execution */	
	return executor_external(cmd, verbose);
}



/**
 * executor_chaser()
 * Executes a command for a chaser
 * @param cmd      Command to execute
 * @param verbose  if zero no output is thrown (MUST BE ZERO FOR THREADS!)
**/
int executor_chaser(int pid, const char *cmd, int val, int verbose)
{
	int d;
	const char *str;
	char measurestr[25];

	/* toogle */
	if(dmxstr_chkcmd(cmd, "toogle") != 0)
	{
		if( !dmxplayer_toogle_program(pid-1) )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		return EXECUTOR_SUCCESS;
	}

	/* start */
	else if(dmxstr_chkcmd(cmd, "on") != 0)
	{
		if( !dmxplayer_start_program(pid-1) )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		return EXECUTOR_SUCCESS;
	}

	/* stop */
	else if(dmxstr_chkcmd(cmd, "off") != 0)
	{
		if( !dmxplayer_stop_program(pid-1) )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		return EXECUTOR_SUCCESS;
	}

	/* faster */
	else if(dmxstr_chkcmd(cmd, "faster") != 0)
	{
		dmxprogram_t *prog;
		prog = dmxplayer_getprogram(pid-1);
		if( prog == 0 )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		/* increase step length */
		dmxprogram_speed_increase( prog );
		/* show new measure */
		dmxstr_translate_steplen(measurestr, prog->steplen);
		if(verbose) cmdbuf_printf("chaser %d measure: %s", pid, measurestr);
		return EXECUTOR_SUCCESS;
	}

	/* slower */
	else if(dmxstr_chkcmd(cmd, "slower") != 0)
	{
		dmxprogram_t *prog;
		prog = dmxplayer_getprogram(pid-1);
		if( prog == 0 )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		/* increase step length */
		dmxprogram_speed_decrease( prog );
		/* show new measure */
		dmxstr_translate_steplen(measurestr, prog->steplen);
		if(verbose) cmdbuf_printf("chaser %d measure: %s", pid, measurestr);
		return EXECUTOR_SUCCESS;
	}

	/* select */
	else if(dmxstr_chkcmd(cmd, "select") != 0)
	{
		dmxprogram_t *prog;
		prog = dmxplayer_getprogram(pid-1);
		if( prog == 0 )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		listing_setSelectedProgram(pid-1);
		if(verbose) cmdbuf_printf("chaser %d selected for record", pid);
		return EXECUTOR_SUCCESS;
	}

	/* set chaser group */
	else if( ((str = dmxstr_chkcmd(cmd, "group")) != 0) || 
	         ((str = dmxstr_chkcmd(cmd, "grp"))   != 0)    )
	{
		/* get parameter */
		if( (str = dmxstr_get_next_param_int(str, &d)) != 0 )
		{
			dmxprogram_t *prog;
			prog = dmxplayer_getprogram(pid-1);
			if( prog == 0 )
			{
				if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
				return EXECUTOR_FAILURE;
			}
			prog->group = d;
			return EXECUTOR_SUCCESS;
		}
		/* no param given */
		if(verbose) cmdbuf_printf("usage: chaser %d group <number>", pid);
		return EXECUTOR_FAILURE;
	}

	/* select */
	else if( (str = dmxstr_chkcmd(cmd, "rename")) != 0 )
	{
		dmxprogram_t *prog;
		prog = dmxplayer_getprogram(pid-1);
		if( prog == 0 )
		{
			if(verbose) cmdbuf_printf("invalid chaser id: %d", pid);
			return EXECUTOR_FAILURE;
		}
		str = dmxstr_remove_whitespace(str);
		dmxprogram_set_name(prog, str);
		if(verbose) cmdbuf_printf("chaser %d renamed to '%s'", pid, str);
		return EXECUTOR_SUCCESS;
	}

	/* command not found */
	cmd = dmxstr_remove_whitespace(cmd);
	if(verbose) cmdbuf_printf("syntax error at: >>%s<<", cmd);
	if(verbose) cmdbuf_printf("usage: chaser %d <command>", pid);
	return EXECUTOR_FAILURE;
}



/**
 * executor_channel()
 * Executes a command for a channel
 * @param cmd      Command to execute
 * @param verbose  if zero no output is thrown (MUST BE ZERO FOR THREADS!)
**/
int executor_channel(int from, int to, const char *cmd, int verbose)
{
	const char *str;
	int d, n;
	int even_option = 0;

	/* check optional even/odd keywords */
	if( (str = dmxstr_chkcmd(cmd, "even")) != 0 )
	{
		even_option = 2;
		cmd = str;
	}
	else if( (str = dmxstr_chkcmd(cmd, "odd")) != 0 )
	{
		even_option = 1;
		cmd = str;
	}

	/* to (set) */
	if( (str = dmxstr_chkcmd(cmd, "at")) != 0)
	{
		/* get parameter */
		if( (str = dmxstr_get_next_param_int(str, &d)) != 0 )
		{
			/* iterate channels */
			for(n=from; n<=to; n++)
			{
				if( (even_option==1) && ((n%2) == 0) ) continue;
				if( (even_option==2) && ((n%2) == 1) ) continue;
				if( dmxplayer_set_prog(n-1, d) == 0 )
				{
					if(verbose) cmdbuf_printf("invalid channel or value");
					return EXECUTOR_FAILURE;
				}
			}
			return EXECUTOR_SUCCESS;
		}
		/* no param given */
		if(verbose) cmdbuf_printf("usage: channel %d to <value>", from);
		return EXECUTOR_FAILURE;
	}

	/* off */
	else if(dmxstr_chkcmd(cmd, "off") != 0)
	{
		/* iterate channels */
		for(n=from; n<=to; n++)
		{
			if( (even_option==1) && ((n%2) == 0) ) continue;
			if( (even_option==2) && ((n%2) == 1) ) continue;
			if( dmxplayer_set_prog(n-1, 0) == 0 )
			{
				if(verbose) cmdbuf_printf("invalid channel or value");
				return EXECUTOR_FAILURE;
			}
		}
		return EXECUTOR_SUCCESS;
	}

	/* on */
	else if(dmxstr_chkcmd(cmd, "on") != 0)
	{
		/* iterate channels */
		for(n=from; n<=to; n++)
		{
			if( (even_option==1) && ((n%2) == 0) ) continue;
			if( (even_option==2) && ((n%2) == 1) ) continue;
			if( dmxplayer_set_prog(n-1, 255) == 0 )
			{
				if(verbose) cmdbuf_printf("invalid channel or value");
				return EXECUTOR_FAILURE;
			}
		}
		return EXECUTOR_SUCCESS;
	}

	/* clear */
	else if(dmxstr_chkcmd(cmd, "clear") != 0)
	{
		/* iterate channels */
		for(n=from; n<=to; n++)
		{
			if( (even_option==1) && ((n%2) == 0) ) continue;
			if( (even_option==2) && ((n%2) == 1) ) continue;
			if( dmxplayer_unset_prog(n-1) == 0 )
			{
				if(verbose) cmdbuf_printf("invalid channel or value");
				return EXECUTOR_FAILURE;
			}
		}
		return EXECUTOR_SUCCESS;
	}

	/* shift left */
	else if(dmxstr_chkcmd(cmd, "shiftl") != 0)
	{
		int first;
		/* check even/odd function */
		if(even_option != 0)
		{
			if(verbose) cmdbuf_printf("even/odd not available for shift!");
			return EXECUTOR_FAILURE;
		}
		/* remind first value */
		first = dmxplayer_get_prog_value(from-1);
		/* iterate channels */
		for(n=from; n<to; n++)
		{
			if( dmxplayer_set_prog(n-1, dmxplayer_get_prog_value(n) ) == 0 )
			{
				if(verbose) cmdbuf_printf("invalid channel or value");
				return EXECUTOR_FAILURE;
			}
		}
		/* set last channel */
		if( dmxplayer_set_prog(to-1, first ) == 0 )
		{
			if(verbose) cmdbuf_printf("invalid channel or value");
			return EXECUTOR_FAILURE;
		}
		return EXECUTOR_SUCCESS;
	}

	/* shift right */
	else if(dmxstr_chkcmd(cmd, "shift") != 0)
	{
		int last;
		/* check even/odd function */
		if(even_option != 0)
		{
			if(verbose) cmdbuf_printf("even/odd not available for shift!");
			return EXECUTOR_FAILURE;
		}
		/* remind first value */
		last = dmxplayer_get_prog_value(to-1);
		/* iterate channels */
		for(n=to-1; n>=from; n--)
		{
			if( dmxplayer_set_prog(n, dmxplayer_get_prog_value(n-1) ) == 0 )
			{
				if(verbose) cmdbuf_printf("invalid channel or value");
				return EXECUTOR_FAILURE;
			}
		}
		/* set last channel */
		if( dmxplayer_set_prog(from-1, last ) == 0 )
		{
			if(verbose) cmdbuf_printf("invalid channel or value");
			return EXECUTOR_FAILURE;
		}
		return EXECUTOR_SUCCESS;
	}

	/* command not found */
	cmd = dmxstr_remove_whitespace(cmd);
	if(verbose) cmdbuf_printf("syntax error at: >>%s<<", cmd);
	return EXECUTOR_FAILURE;
}



/**
 * executor_exec()
 * Executes a script stored in a file
 * @param cmd      Command to execute
 * @param verbose  if zero no output is thrown (MUST BE ZERO FOR THREADS!)
**/
int executor_exec(const char *file, int verbose)
{
	char buf[512];
	int line=0;
	/* open file */
	FILE *rf = fopen(file, "r");
	/* check file */
	if(rf == NULL) 
	{
		if(verbose) cmdbuf_printf("script file '%s' not found", file);
		return EXECUTOR_FAILURE;
	}
	/* read the hole file */
	while(fgets(buf, sizeof(buf), rf))
	{
		line++;
		buf[sizeof(buf)-1] = '\0';
		if( executor(buf, verbose) != EXECUTOR_SUCCESS )
		{
			if(verbose) cmdbuf_printf("%s: Error on line %d", file, line);
			break;
		}
	}	
	/* close file */
	fclose(rf);
	return EXECUTOR_SUCCESS;
}



/**
 * executor_external()
 * Executes an external command (in a script in binary dir)
 * @param cmd      Command to execute
 * @param verbose  if zero no output is thrown (MUST BE ZERO FOR THREADS!)
**/
int executor_external(const char *cmd, int verbose)
{
	int ret;
	struct stat fileStat;
	char filebuf[EXECUTOR_FILENAME_BUFLEN];

	/* check script directory */
	if(stat(EXECUTOR_SCRIPTDIR, &fileStat) < 0)
	{
		/* command not found */
		if(verbose)
		{
			cmdbuf_printf("command not known ("
			              "can not open script directory '%s'!)", 
			              EXECUTOR_SCRIPTDIR);
		}
		return EXECUTOR_NOTFOUND;
	}
	if( !(S_ISDIR(fileStat.st_mode)) ) 
	{
		/* command not found */
		if(verbose)
		{
			cmdbuf_printf("command not known ("
			              "script directory '%s' does not seem to be a directory!)", 
			              EXECUTOR_SCRIPTDIR);
		}
		return EXECUTOR_NOTFOUND;
	}

	/* generate filename */
	ret = dmxstr_get_script_filename(filebuf, 
	                                 EXECUTOR_FILENAME_BUFLEN,
	                                 EXECUTOR_SCRIPTDIR,
	                                 cmd);
	if(ret == 0)
	{
		/* command not found */
		if(verbose)
		{
			cmdbuf_printf("command not known ("
			              "filename buffer overflow!)", 
			              EXECUTOR_SCRIPTDIR);
		}
		return EXECUTOR_NOTFOUND;
	}

	/* check file */
	if(stat(filebuf, &fileStat) < 0)
	{
		/* command not found */
		if(verbose)
		{
			cmdbuf_printf("command not known", 
			              filebuf);
		}
		return EXECUTOR_NOTFOUND;
	}
	if( (S_ISDIR(fileStat.st_mode)) ) 
	{
		/* command not found */
		if(verbose)
		{
			cmdbuf_printf("command not known ("
			              "script for command '%s' is a directory!)", 
			              filebuf);
		}
		return EXECUTOR_NOTFOUND;
	}

	/* EXECUTE */
	executor_exec(filebuf, verbose);

	/* check if exists */
	return EXECUTOR_SUCCESS;
}



/**
 * executor_midibind()
 * Binds MIDI to a command
 * @param cmd      Command to execute
 * @param verbose  if zero no output is thrown (MUST BE ZERO FOR THREADS!)
**/
int executor_midibind(const char *cmd, int verbose)
{
	int b1, b2;
	const char *str;
	if( (str = dmxstr_chkcmd(cmd, ":")) != 0 )
	{
		str = dmxstr_remove_whitespace(str);
		/* add translation */
		midireceiver_add_translation_l(str);
		if(verbose) cmdbuf_printf("bound learned value to %s", str);
		return EXECUTOR_SUCCESS;
	}
	/* get parameter 1 */
	else if( (str = dmxstr_get_next_param_int(cmd, &b1)) != 0 )
	{
		/* get parameter 2 */
		if( (str = dmxstr_get_next_param_int(str, &b2)) != 0 )
		{
			if( (str = dmxstr_chkcmd(str, ":")) != 0 )
			{
				str = dmxstr_remove_whitespace(str);
				/* add translation */
				midireceiver_add_translation(b1, b2, str);
				if(verbose) cmdbuf_printf("bound (%d,%d) to %s", b1, b2, str);
				return EXECUTOR_SUCCESS;
			}
		}
	}
	/* no param given */
	if(verbose) cmdbuf_printf("usage: midibind <byte1> <byte2> : <command>");
	if(verbose) cmdbuf_printf("[or]   midibind : <command>");
	return EXECUTOR_FAILURE;
}
