/**
 * @file lindxmd_config.c
 * @author Florian Schmitt <florian.schmitt@qikgrp.com>
 * ---------------------------------------------------------
 * (c) Florian Schmitt / 2012-2014 / All rights reserved. 
 * ---------------------------------------------------------
 * This provides config access.
**/
#include "lindmxd_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#define CONF_ERR_BUFLEN   200
char conferr[CONF_ERR_BUFLEN];

/**
 * lindmxd_config_init_defaults()
 * Sets default values to all config parameters
 * @param conf    The config structure to initialize
**/
void lindmxd_config_init_defaults(lindmxd_config_t *conf)
{
	if( conf == NULL ) return;
	conf->program_name = "lindmxd";
	conf->daemon_name  = "lindmxd";
	conf->channels     = 100;
	conf->port         = 6767;
	conf->debug        = 0;
	conf->zero         = 0;
	conf->log_upto     = LOG_NOTICE;
	conf->log_facility = LOG_DAEMON;
	/* interface options */
	conf->interface    = "dmx4all_micro_usb";
	conf->if_port      = 10001;
	strncpy(conf->if_host, "192.168.0.250", LINDMXD_CONFIG_BUFLEN);
}



/**
 * Reads from file linewise and builds a string ending on \0
**/
int config_fgets(char *buf, size_t buflen, FILE *fp)
{
	int c;
	int n=0;
	while ( ((c = fgetc(fp)) != EOF) &&
	         (c != '\n') && (c != '\r') &&
	         (n < (buflen-1)) )
		buf[n++] = c;
	buf[n] = '\0';
	if(c < 0) return c;
	return n;
}



/**
 * Checks if a string is either yes/no, true/false, 1/0
 * Returns:
 *   1 if yes
 *   0 if no
 *  -1 if invalid answer
**/
int parse_yesno(const char *str)
{
	if( strcasecmp("1",     str) == 0 ) return 1;
	if( strcasecmp("yes",   str) == 0 ) return 1;
	if( strcasecmp("true",  str) == 0 ) return 1;
	if( strcasecmp("0",     str) == 0 ) return 0;
	if( strcasecmp("no",    str) == 0 ) return 0;
	if( strcasecmp("false", str) == 0 ) return 0;
	return -1;
}



/**
 * lindmxd_config_parse()
 * Loads the configuration from a file
 *  DO NOT CALL FROM THREADS (!)
 * @param conf       The config structure
 * @param filename   The file to load from
 * @param maxrek     Maximum rekursion / include level
 * @return 1 on success, -1 on parser error, -2 on file not found,
 *         -3 on internal error, -4 on rekursion exceed
**/
int lindmxd_config_parse(lindmxd_config_t *conf, const char *filename, int maxrek)
{
	int retval;
	int line = 0;
	char *pch;
	const char *whitespace = " \t";
	char linebuf[LINDMXD_CONFIG_BUFLEN];
	if( filename == NULL ) return LINDMXD_CONFIG_ERR_FILE;
	if( conf     == NULL ) return LINDMXD_CONFIG_ERR_INTERNAL;
	if( maxrek   <= 0 )
	{
		snprintf(conferr, CONF_ERR_BUFLEN, 
		         "maximum include level exceeded by including %s", filename);
		return LINDMXD_CONFIG_ERR_MAXREK;
	}

	/* Open config for reading */
	FILE *tf = fopen(filename, "r");
	if(!tf) 
	{
		snprintf(conferr, CONF_ERR_BUFLEN, "can not open config file %s", filename);
		return LINDMXD_CONFIG_ERR_FILE;
	}

	while( config_fgets(linebuf, LINDMXD_CONFIG_BUFLEN, tf) >= 0 )
	{
		line++;
		if(linebuf[0] == '\0') continue;
		pch = strtok(linebuf, whitespace);
		/* Check comment / empty */
		if( (pch[0] == '#') || (pch[0] == '\0') ) continue;
		/* command: debug */
		if( strcasecmp("debug" , pch) == 0 )
		{
			/* yes or no */
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			conf->debug = parse_yesno(pch);
			if(conf->debug == -1) goto out_err;
			if(conf->debug ==  1) conf->log_upto = LOG_DEBUG;
			/* check end */
			if( strtok(NULL, whitespace) != NULL )
				goto out_err;
		}
		/* command: zero */
		else if( strcasecmp("zero" , pch) == 0 )
		{
			/* yes or no */
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			conf->zero = parse_yesno(pch);
			if(conf->zero == -1) goto out_err;
			/* check end */
			if( strtok(NULL, whitespace) != NULL )
				goto out_err;
		}
		/* command: channels */
		else if( strcasecmp("channels" , pch) == 0 )
		{
			/* yes or no */
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			conf->channels = atoi(pch);
			if(conf->channels < 0) goto out_err;
			/* check end */
			if( strtok(NULL, whitespace) != NULL )
				goto out_err;
		}
		/* command: include */
		else if( strcasecmp("include" , pch) == 0 )
		{
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			/* check end */
			if( strtok(NULL, whitespace) != NULL )
				goto out_err;
			/* rekursive parsing */
			retval = lindmxd_config_parse(conf, pch, --maxrek);
			if( retval != LINDMXD_CONFIG_SUCCESS )
			{
				fclose(tf);
				return retval;
			}
		}
		/* command: listen */
		else if( strcasecmp("listen" , pch) == 0 )
		{
			/* get protocol */
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			if( strcasecmp("tcp" , pch) == 0 )
				fprintf(stderr, "protocol selection not implemented\n");
			else if( strcasecmp("udp" , pch) == 0 )
				fprintf(stderr, "protocol selection not implemented\n");
			else goto out_err;
			/* get port */
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			conf->port = atoi( pch );
			/* check end */
			if( strtok(NULL, whitespace) != NULL )
				goto out_err;
		}
		/* command: interface */
		else if( strcasecmp("interface" , pch) == 0 )
		{
			/* get if_type */
			pch = strtok(NULL, whitespace);
			if(pch == NULL) goto out_err;
			if( strcasecmp("dmx4all_micro_usb" , pch) == 0 )
				conf->interface = "dmx4all_micro_usb";
			else if( strcasecmp("simulate" , pch) == 0 )
				conf->interface = "simulate";
			else if(strcasecmp("dmx4all_tcp" , pch) == 0)
			{
				conf->interface = "dmx4all_tcp";
				/* get if_addr */
				pch = strtok(NULL, whitespace);
				if(pch != NULL)
				{
					strncpy(conf->if_host, pch, LINDMXD_CONFIG_BUFLEN);
					/* get if_port */
					pch = strtok(NULL, whitespace);
					if(pch != NULL)
						conf->if_port = atoi( pch );
				}
			}
			else if(strcasecmp("dmx4all_udp" , pch) == 0)
			{
				conf->interface = "dmx4all_udp";
				/* get if_addr */
				pch = strtok(NULL, whitespace);
				if(pch != NULL)
				{
					strncpy(conf->if_host, pch, LINDMXD_CONFIG_BUFLEN);
					/* get if_port */
					pch = strtok(NULL, whitespace);
					if(pch != NULL)
						conf->if_port = atoi( pch );
				}
			}
			else goto out_err;
			/* check end */
			if( (pch != NULL) && (strtok(NULL, whitespace) != NULL) )
				goto out_err;
		}
		/* invalid command */
		else goto out_err;
	}

	/* close config file */
	fclose(tf);
	return LINDMXD_CONFIG_SUCCESS;
out_err:
	fclose(tf);
	snprintf(conferr, CONF_ERR_BUFLEN, "error parsing config (%s) on line %d", filename, line);
	return LINDMXD_CONFIG_ERR_PARSE;
}



/**
 * lindmxd_config_geterr()
 * Converts error code to string
 * @param error    Error code
 * @return String, describing the error code
**/
const char * lindmxd_config_geterr()
{
	return conferr;
}
