/**
 * @file modules/dmx4all_udp.c
 * @author Florian Schmitt <florian.schmitt@qikgrp.com>
 * ----------------------------------------------------------------------------
 * Implements dmx4all network interface driver
 * ----------------------------------------------------------------------------
 * (c) Florian Schmitt / 2012-2013
 * License GPLv3+: GNU GPL Version 3
 * or higher <http://gnu.org/licenses/gpl.html>
 * This code comes with ABSOLUTELY NO WARRANTY. This is free software, and 
 * you are welcome to redistribute it under certain conditions.
 * ----------------------------------------------------------------------------
 * This module is sponsored by zonezero GmbH <http://www.zonezero.de>
 * ----------------------------------------------------------------------------
**/
#include "dmx4all_udp.h"
#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define BUFLEN 80

/* globals */
int d4a_udp_maxchn       = 0;
const char *d4a_udp_host = 0;
int d4a_udp_port         = 0;
int d4a_udp_sock         = 0;
struct sockaddr_in d4a_udp_addr;



/**
 * dmx4all_udp_connect()
 * Reconnects to interface
 * @return 1 on success, 0 on failure
**/
int dmx4all_udp_connect()
{
	char buffer[BUFLEN];
	struct hostent *host_ent;

	/* check settings */
	if(d4a_udp_host == NULL)
	{
		syslog(LOG_ERR, "Illegal host value for networking interface.");
		return 0;
	}
	if(d4a_udp_port == 0)
	{
		syslog(LOG_ERR, "Illegal port value for networking interface.");
		return 0;
	}

	/* resolve hostname */
	if ( (host_ent = gethostbyname(d4a_udp_host)) == NULL) 
	{
		syslog(LOG_ERR, "Interface host'%s' can not be resolved", d4a_udp_host);
		return 0;
	}

	/* create socket */
	d4a_udp_sock = socket(AF_INET, SOCK_DGRAM, 0); // IPPROTO_UDP

	/* set server address */
	bzero(&d4a_udp_addr, sizeof(d4a_udp_addr));
	d4a_udp_addr.sin_family       = AF_INET;
	d4a_udp_addr.sin_addr.s_addr  = *((unsigned long *) host_ent->h_addr_list[0]);
	d4a_udp_addr.sin_port         = htons(d4a_udp_port);

	/* send data */
	sprintf(buffer, "N%03d\r\n", d4a_udp_maxchn);
	sendto(d4a_udp_sock, buffer, strlen(buffer), 0,
	       (struct sockaddr *)&d4a_udp_addr,
	       sizeof(d4a_udp_addr));

	return 1;
}



/**
 * dmx4all_udp_init()
 * Init DMX4ALL Micro USB Interface
 * @param maxchannel    The highest channel number created
 * @return 1 on success, 0 on failure
**/
int dmx4all_udp_init(int maxchannel, const char *if_host, int if_port)
{
	/* settings */
	d4a_udp_maxchn  = maxchannel;
	d4a_udp_host    = if_host;
	d4a_udp_port    = if_port;
	/* try connect */
	return dmx4all_udp_connect();
}



/**
 * dmx4all_udp_destroy()
 * Clean up interface
**/
void dmx4all_udp_destroy()
{
	/* close connection */
	close(d4a_udp_sock);
	d4a_udp_sock = 0;
}



/**
 * dmx4all_udp_reset_channels()
 * Resets all channels to zero
 * @param maxchannel    The highest channel number created
 * @return 1 on success, 0 on failure
**/
int dmx4all_udp_reset_channels()
{
	return dmx4all_udp_set_all( 0 );
}



/**
 * dmx4all_udp_set_all()
 * Sets all channels to a given value
 * @param value    Value to set all channels to
 * @return 1 on success, 0 on failure
**/
int dmx4all_udp_set_all(int value)
{
	int n, ret;
	/* loop through all channels */
	ret = 1;
	for(n=0; n<d4a_udp_maxchn; n++)
	{
		if(!dmx4all_udp_set(n, value)) ret = 0;
	}
	return ret;
}



/**
 * dmx4all_udp_set()
 * Sets a channel to a given value
 * @param channel  Channel to set
 * @param value    Value to set to
 * @return 1 on success, 0 on failure
**/
int dmx4all_udp_set(int channel, int value)
{
	char buffer[BUFLEN];

	/* send data */
	sprintf(buffer, "C%03dL%03d\r\n", channel, value);
	sendto(d4a_udp_sock, buffer, strlen(buffer), 0,
	       (struct sockaddr *)&d4a_udp_addr,
	       sizeof(d4a_udp_addr));

	/* success */
	return 1;
}
